# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' System editor interaction. '''


from __future__ import annotations

from . import __


_scribe = __.produce_scribe( __name__ )


def discover_editor( ) -> __.cabc.Callable[ [ str ], str ]:
    ''' Discovers editor and returns executor function. '''
    from shutil import which
    from subprocess import run # nosec B404
    editor = (
            __.os.environ.get( 'VISUAL' )
        or  __.os.environ.get( 'EDITOR' )
        # TODO: Better default for Windows.
        or  'nano' )
    # TODO: Platform-specific list.
    for editor_ in ( editor, 'nano' ):
        if which( editor_ ):
            editor = editor_
            break
    else: editor = ''

    if editor:

        # TODO? async
        def editor_executor( filename: str ) -> str:
            ''' Executes editor with file. '''
            run( ( editor, filename ), check = True ) # nosec B603
            with open( filename, 'r', encoding = 'utf-8' ) as stream:
                return stream.read( )

        return editor_executor

    from .exceptions import ProgramAbsenceError
    raise ProgramAbsenceError( 'editor' )


def edit_content(
    content: str = '', *,
    suffix: str = '.md',
    editor_discoverer: __.cabc.Callable[
        [ ], __.cabc.Callable[ [ str ], str ] ] = discover_editor,
) -> str:
    ''' Edits content via discovered editor. '''
    from .exceptions import EditorFailure, ProgramAbsenceError
    try: editor = editor_discoverer( )
    except ProgramAbsenceError:
        _scribe.exception( "Could not find editor program." )
        return content
    import tempfile
    with tempfile.NamedTemporaryFile( mode = 'r+', suffix = suffix ) as tmp:
        tmp.write( content )
        tmp.flush( )
        try: return editor( tmp.name )
        except Exception as exc: raise EditorFailure( cause = exc ) from exc



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Preparation of the library core. '''


from . import imports as __
from . import application as _application
from . import configuration as _configuration
from . import dictedits as _dictedits
from . import distribution as _distribution
from . import environment as _environment
from . import inscription as _inscription
from . import state as _state


async def prepare( # pylint: disable=too-many-arguments,too-many-locals
    exits: __.ExitsAsync,
    application: _application.Information = _application.Information( ),
    configedits: _dictedits.Edits = ( ),
    configfile: __.Absential[ __.Path ] = __.absent,
    environment: bool = False,
    inscription: __.Absential[ _inscription.Control ] = __.absent,
) -> _state.Globals:
    ''' Prepares globals DTO for use with library functions.

        Also:
        * Configures logging for library package (not application).
        * Optionally, loads process environment from files.

        Note that asynchronous preparation allows for applications to
        concurrently initialize other entities outside of the library, even
        though the library initialization, itself, is inherently sequential.
    '''
    directories = application.produce_platform_directories( )
    distribution = (
        await _distribution.Information.prepare(
            package = __.package_name, exits = exits ) )
    configuration = (
        await _configuration.acquire(
            application_name = application.name,
            directories = directories,
            distribution = distribution,
            edits = configedits,
            file = configfile ) )
    auxdata = _state.Globals(
        application = application,
        configuration = configuration,
        directories = directories,
        distribution = distribution,
        exits = exits )
    if environment: await _environment.update( auxdata )
    if __.is_absent( inscription ):
        inscription_: _inscription.Control = _inscription.Control( )
    else: inscription_ = inscription
    _inscription.prepare( control = inscription_ )
    _inscribe_preparation_report( auxdata )
    return auxdata


def _inscribe_preparation_report( auxdata: _state.Globals ):
    scribe = __.produce_scribe( __.package_name )
    scribe.debug( f"Application Name: {auxdata.application.name}" )
    # scribe.debug( f"Execution ID: {auxdata.application.execution_id}" )
    scribe.debug( "Application Cache Location: {}".format(
        auxdata.provide_cache_location( ) ) )
    scribe.debug( "Application Data Location: {}".format(
        auxdata.provide_data_location( ) ) )
    scribe.debug( "Application State Location: {}".format(
        auxdata.provide_state_location( ) ) )
    scribe.debug( "Package Data Location: {}".format(
        auxdata.distribution.provide_data_location( ) ) )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Information about package distribution. '''


from . import imports as __
from . import io as _io


class Information(
    metaclass = __.ImmutableStandardDataclass,
    decorators = ( __.standard_dataclass, ),
):
    ''' Information about a package distribution. '''

    name: str
    location: __.Path
    editable: bool

    @classmethod
    async def prepare(
        selfclass, package: str, exits: __.ExitsAsync,
        project_anchor: __.Absential[ __.Path ] = __.absent,
    ) -> __.typx.Self:
        ''' Acquires information about our package distribution. '''
        import sys
        # Detect PyInstaller bundle.
        if getattr( sys, 'frozen', False ) and hasattr( sys, '_MEIPASS' ):
            project_anchor = __.Path( getattr( sys, '_MEIPASS' ) )
        from importlib.metadata import packages_distributions
        # https://github.com/pypa/packaging-problems/issues/609
        name = packages_distributions( ).get( package )
        if name is None: # Development sources rather than distribution.
            editable = True # Implies no use of importlib.resources.
            location, name = (
                await _acquire_development_information(
                    location = project_anchor ) )
        else:
            editable = False
            name = name[ 0 ]
            location = await _acquire_production_location( package, exits )
        return selfclass(
            editable = editable, location = location, name = name )

    def provide_data_location( self, *appendages: str ) -> __.Path:
        ''' Provides location of distribution data. '''
        base = self.location / 'data'
        if appendages: return base.joinpath( *appendages )
        return base


async def _acquire_development_information(
    location: __.Absential[ __.Path ] = __.absent
) -> tuple[ __.Path, str ]:
    from tomli import loads
    if __.is_absent( location ):
        location = __.Path( __file__ ).parents[ 3 ].resolve( strict = True )
    pyproject = await _io.acquire_text_file_async(
        location / 'pyproject.toml', deserializer = loads )
    name = pyproject[ 'project' ][ 'name' ]
    return location, name


async def _acquire_production_location(
    package: str, exits: __.ExitsAsync
) -> __.Path:
    # TODO: Python 3.12: importlib.resources
    # TODO: 'importlib_resources' PR to fix 'as_file' type signature.
    from importlib_resources import files, as_file # pyright: ignore
    # Extract package contents to temporary directory, if necessary.
    return exits.enter_context(
        as_file( files( package ) ) ) # pyright: ignore



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Sensitive filesystem locations on Windows. '''


from __future__ import annotations

from . import __


_scribe = __.produce_scribe( __name__ )


CSIDL_PROGRAM_FILES_COMMON = 0x002b    # C:\Program Files\Common Files
CSIDL_PROGRAM_FILES_COMMONX86 = 0x002c # C:\Program Files (x86)\Common Files


def discover_system_paths( ) -> frozenset[ __.Path ]:
    ''' Discovers system paths via standard Windows mechanisms. '''
    paths: set[ __.Path ] = set( )
    paths.update( _discover_system_paths_via_environment( ) )
    paths.update( _discover_system_paths_via_api( ) )
    # TODO? Cygwin
    if _detect_mingw( ): _discover_add_mingw_system_paths( paths )
    return frozenset( paths )


def _detect_mingw( ) -> bool:
    ''' Checks if running in MinGW environment. '''
    # TODO: If environment vars are not set, then return False.
    #       Else, proceed to more expensive checck to ensure sane MinGW.
    mingw_env = __.os.environ.get( 'MSYSTEM', '' )
    if mingw_env.startswith( ('MINGW', 'MSYS') ):
        _scribe.debug( f'MinGW detected via MSYSTEM={mingw_env}' )
        return True
    mingw_paths = ( '/mingw32', '/mingw64', '/msys', '/usr/bin/msys-2.0.dll' )
    for path in mingw_paths:
        if not __.Path( path ).exists( ): continue
        _scribe.debug( f'MinGW detected via path: {path}' )
        return True
    return False


def _discover_add_mingw_system_paths( paths: set[ __.Path ] ) -> None:
    paths_: set[ __.Path ] = set( )
    for path in paths:
        parts = path.parts
        if 1 >= len( parts ): continue
        drive = parts[ 0 ][ 0 ].lower( ) # TODO? Consider UNC paths.
        paths_.add( __.Path( f"/{drive}" ).joinpath( *parts[ 1 : ] ) )
    _scribe.debug(
        "Calculated {} MingGW system paths.".format( len( paths_ ) ) )
    paths_.update( map(
        __.Path,
        (   '/bin', '/dev', '/etc',
            '/mingw32', '/mingw64', '/msys', '/proc', '/usr',
        ) ) )
    paths.update( paths_ )


def _discover_system_paths_via_environment( ) -> frozenset[ __.Path ]:
    ''' Discovers system paths via environment. '''
    environ = __.os.environ
    paths: set[ __.Path ] = set( )
    system_drive = environ.get( 'SystemDrive', 'C:' )
    system_root = environ.get( 'SystemRoot', f"{system_drive}\\Windows" )
    paths.add( __.Path( system_root ) )
    paths.add( __.Path( f"{system_drive}/System Volume Information" ) )
    # Program Files variations
    progfiles = __.os.environ.get(
        'ProgramFiles', f"{system_drive}\\Program Files" )
    paths.add( __.Path( progfiles ) )
    for progfiles_ename in (
        'ProgramFiles(x86)', 'ProgramFiles(Arm)', 'ProgramW6432'
    ):
        progfiles = environ.get( progfiles_ename )
        if progfiles: paths.add( __.Path( progfiles ) )
    return frozenset( paths )


def _discover_system_paths_via_api( ) -> set[ __.Path ]:
    ''' Discovers system paths via API. '''
    import ctypes
    from ctypes.wintypes import MAX_PATH
    dmessage = "Could not retrieve additional Windows system paths via API."
    paths: set[ __.Path ] = set( )
    try: dll = ctypes.windll.shell32 # pyright: ignore
    except Exception: # pylint: disable=broad-exception-caught
        _scribe.debug( dmessage )
    buf = ctypes.create_unicode_buffer( MAX_PATH + 1 )
    for key in (
        CSIDL_PROGRAM_FILES_COMMON,
        CSIDL_PROGRAM_FILES_COMMONX86,
    ):
        try:
            dll.SHGetFolderPathW( # pyright: ignore
                None, key, None, 0, buf )
            paths.add( __.Path( buf.value ) )
        except Exception: # pylint: disable=broad-exception-caught
            _scribe.debug( dmessage )
    return paths


def discover_user_paths( ) -> frozenset[ __.Path ]:
    ''' Discovers Windows user-specific paths that should be protected. '''
    paths: set[ __.Path ] = set( )
    appdata = __.os.environ.get( 'APPDATA' )
    local_appdata = __.os.environ.get( 'LOCALAPPDATA' )
    userprofile = __.os.environ.get( 'USERPROFILE' )
    if appdata: paths.add( __.Path( appdata ) )
    if local_appdata: paths.add( __.Path( local_appdata ) )
    if userprofile:
        profile = __.Path( userprofile )
        paths.update( (
            profile / 'AppData' / 'Local',
            profile / 'AppData' / 'LocalLow',
            profile / 'AppData' / 'Roaming',
            # Legacy Paths
            profile / 'Application Data',
            profile / 'Local Settings',
        ) )
    return frozenset( paths )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Sensitive filesystem locations relative to project directory. '''


from __future__ import annotations

from . import __


_scribe = __.produce_scribe( __name__ )


def discover_sensitive_locations( ) -> frozenset[ str ]:
    ''' Discovers sensitive directories in project context. '''
    return frozenset( (
        # Version Control
        '.git', '.svn', '.hg', '.bzr',
        # Project Config
        '.idea', '.vscode', '.eclipse',
        # Build and Dependencies
        'node_modules', '.virtualenv', '.env',
        # Infrastructure
        '.terraform', '.ansible',
        # Secrets
        '.secrets', '.env',
    ) )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Abstract bases and interfaces. '''


from __future__ import annotations

from . import __
from . import parts as _parts


class CliCommand( # pylint: disable=invalid-metaclass
    __.typx.Protocol,
    metaclass = __.ImmutableStandardProtocolDataclass,
    decorators = ( __.standard_dataclass, __.typx.runtime_checkable ),
):
    ''' CLI command. '''

    @__.abc.abstractmethod
    async def __call__( self, auxdata: __.Globals ) -> None:
        ''' Executes command with global state. '''
        raise NotImplementedError

    @__.abc.abstractmethod
    def provide_configuration_edits( self ) -> __.DictionaryEdits:
        ''' Provides edits against configuration from options. '''
        raise NotImplementedError


class DifferencesDisplay( # pylint: disable=invalid-metaclass
    __.typx.Protocol,
    metaclass = __.ImmutableStandardProtocolDataclass,
    decorators = ( __.standard_dataclass, __.typx.runtime_checkable ),
):
    ''' Configuration for content differences display. '''

    context_lines: int = 3
    inline_threshold: int = 24 # TODO? Adjust to terminal height.
    # TODO: colorize: bool = False
    # TODO: emojify: bool = False
    # TODO: show_whitespace: bool = False
    # TODO: truncate_lines: bool = False
    # TODO: wrap_lines: bool = False

    @__.abc.abstractmethod
    async def __call__( self, lines: __.cabc.Sequence[ str ] ) -> None:
        ''' Displays differences segment. '''
        raise NotImplementedError


class DifferencesInteractor( # pylint: disable=invalid-metaclass
    __.typx.Protocol,
    metaclass = __.ImmutableProtocolClass,
):
    ''' Interactions with content differences. '''

    @__.abc.abstractmethod
    async def __call__(
        self, lines: __.cabc.Sequence[ str ], display: DifferencesDisplay
    ) -> bool:
        ''' Prompts for action on differences segment. '''
        raise NotImplementedError


class PartInteractor( # pylint: disable=invalid-metaclass
    __.typx.Protocol,
    metaclass = __.ImmutableStandardProtocolDataclass,
    decorators = ( __.standard_dataclass, __.typx.runtime_checkable ),
):
    ''' Interactions with mimeogram parts. '''

    @__.abc.abstractmethod
    async def __call__(
        self, target: _parts.Target
    ) -> tuple[ _parts.Resolutions, str ]:
        ''' Prompts for action on mimeogram part. '''
        raise NotImplementedError



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Mimeogram prompt text for LLMs. '''
# TODO? Use BSD sysexits.


from __future__ import annotations

from . import __
from . import interfaces as _interfaces


_scribe = __.produce_scribe( __name__ )


class Command(
    _interfaces.CliCommand,
    decorators = ( __.standard_dataclass, __.standard_tyro_class ),
):
    ''' Provides LLM prompt text for mimeogram format. '''

    clip: __.typx.Annotated[
        __.typx.Optional[ bool ],
        __.tyro.conf.arg(
            aliases = ( '--clipboard', '--to-clipboard' ),
            help = "Copy prompt to clipboard." ),
    ] = None

    async def __call__( self, auxdata: __.Globals ) -> None:
        ''' Executes command to provide prompt text. '''
        await provide_prompt( auxdata )

    def provide_configuration_edits( self ) -> __.DictionaryEdits:
        ''' Provides edits against configuration from options. '''
        edits: list[ __.DictionaryEdit ] = [ ]
        if None is not self.clip:
            edits.append( __.SimpleDictionaryEdit( # pyright: ignore
                address = ( 'prompt', 'to-clipboard' ), value = self.clip ) )
        return tuple( edits )


async def acquire_prompt( auxdata: __.Globals ) -> str:
    ''' Acquires prompt text from package data. '''
    location = (
        auxdata.distribution.provide_data_location(
            'prompts', 'mimeogram.md' ) )
    return await __.acquire_text_file_async( location )


async def provide_prompt( auxdata: __.Globals ) -> None:
    ''' Provides mimeogram prompt text. '''
    try: prompt = await acquire_prompt( auxdata )
    except Exception as exc:
        _scribe.exception( "Could not acquire prompt text." )
        raise SystemExit( 1 ) from exc
    options = auxdata.configuration.get( 'prompt', { } )
    if options.get( 'to-clipboard', False ):
        from pyperclip import copy
        try: copy( prompt )
        except Exception as exc:
            _scribe.exception( "Could not copy prompt to clipboard." )
            raise SystemExit( 1 ) from exc
        _scribe.info( "Copied prompt to clipboard." )
    else: print( prompt ) # TODO? Use output stream from configuration.
    raise SystemExit( 0 )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Cache for filesystem protection checks. '''


from __future__ import annotations

from . import __
from . import core as _core


_scribe = __.produce_scribe( __name__ )


class Rule(
    metaclass = __.ImmutableStandardDataclass,
    decorators = ( __.standard_dataclass, ),
):
    ''' Rule for path protection. '''

    paths: frozenset[ __.Path ]
    patterns: frozenset[ str ] = frozenset( )


class Cache( _core.Protector, decorators = ( __.standard_dataclass, ) ):
    ''' Cache of protected paths and patterns for platform. '''

    rules: dict[ _core.Reasons, Rule ]
    defaults_disablement: frozenset[ str ]
    rules_supercession: __.ImmutableDictionary[
        __.Path, tuple[ frozenset[ str ], frozenset[ str ] ] ]

    @classmethod
    def from_configuration( selfclass, auxdata: __.Globals ) -> Cache:
        ''' Initializes protection cache for current platform. '''
        _scribe.debug( 'Initializing protection cache.' )
        rules: dict[ _core.Reasons, Rule ] = { }
        discover_platform_locations( auxdata, rules )
        provide_credentials_locations( rules )
        provide_project_locations( rules )
        disables, supercedes = _process_configuration( auxdata, rules )
        return selfclass(
            rules = rules,
            defaults_disablement = disables,
            rules_supercession = supercedes )

    def verify( self, path: __.Path ) -> _core.Status:
        ''' Verifies if a path should be protected using cached data. '''
        path = path.resolve( )
        _scribe.debug( f"Path: {path}" )
        if any( part in self.defaults_disablement for part in path.parts ):
            return _core.Status( path = path, active = False )
        for dir_path, ( ignore, protect ) in self.rules_supercession.items( ):
            if not path.is_relative_to( dir_path ): continue
            rel_path = path.relative_to( dir_path )
            if _check_path_patterns( rel_path, ignore ):
                return _core.Status( path = path, active = False )
            if _check_path_patterns( rel_path, protect ):
                return _core.Status(
                    path = path,
                    reason = _core.Reasons.PlatformSensitive,
                    active = True )
        for reason, rule in self.rules.items( ):
            for protected_path in rule.paths:
                if path.is_relative_to( protected_path ):
                    return _core.Status(
                        path = path,
                        reason = reason,
                        active = True )
            if _check_path_patterns( path, rule.patterns ):
                return _core.Status(
                    path = path,
                    reason = reason,
                    active = True )
        return _core.Status( path = path, active = False )


def provide_credentials_locations(
    rules: dict[ _core.Reasons, Rule ]
) -> None:
    ''' Provides common locations for credentials and other secrets. '''
    from . import home as module
    home = __.Path.home( )
    cred_paths = {
        home / path for path in module.discover_sensitive_locations( ) }
    rules[ _core.Reasons.Credentials ] = Rule(
        paths = frozenset( cred_paths ) )


def provide_project_locations(
    rules: dict[ _core.Reasons, Rule ]
) -> None:
    ''' Provides IDE and VCS locations relative to project. '''
    from . import project as module
    project_sensitive = module.discover_sensitive_locations( )
    # TODO: Consider whether these patterns are compatible with Windows.
    rules[ _core.Reasons.Concealment ] = Rule(
        paths = frozenset( ),
        patterns = frozenset(
            f"**/{path}/**" for path in project_sensitive ) )


def _check_path_patterns( path: __.Path, patterns: frozenset[ str ] ) -> bool:
    ''' Checks if path matches any of the glob patterns. '''
    from wcmatch import glob
    str_path = str( path )
    return glob.globmatch( str_path, list( patterns ), flags = glob.GLOBSTAR )


def discover_platform_locations(
    auxdata: __.Globals, rules: dict[ _core.Reasons, Rule ]
) -> None:
    ''' Discovers system and user locations based on platform. '''
    match __.sys.platform:
        case 'darwin':
            from . import macos as module
            sys_paths = module.discover_system_paths( )
            user_paths = module.discover_user_paths( )
        case 'win32':
            from . import windows as module
            sys_paths = module.discover_system_paths( )
            user_paths = module.discover_user_paths( )
        case _:
            from . import unix as module
            sys_paths = module.discover_system_paths( )
            user_paths: frozenset[ __.Path ] = frozenset( )
    rules[ _core.Reasons.OsDirectory ] = Rule( paths = frozenset( sys_paths ) )
    config_paths = (
        user_paths | { __.Path( auxdata.directories.user_config_path ), } )
    rules[ _core.Reasons.UserConfiguration ] = Rule(
        paths = frozenset( config_paths ) )


def _expand_location( path: str ) -> __.Path:
    ''' Expands path with home directory and environment variables. '''
    expanded = __.os.path.expanduser( __.os.path.expandvars( path ) )
    return __.Path( expanded )


def _process_configuration(
    auxdata: __.Globals,
    rules: dict[ _core.Reasons, Rule ],
) -> tuple[
    frozenset[ str ],
    __.ImmutableDictionary[
        __.Path, tuple[ frozenset[ str ], frozenset[ str ] ] ],
]:
    config = auxdata.configuration.get( 'protection', { } )
    if not config: return frozenset( ), __.ImmutableDictionary( )
    # Additional locations and patterns.
    locations_add = {
        _expand_location( path )
        for path in config.get( 'additional-locations', ( ) ) }
    patterns_add = set( config.get( 'additional-patterns', ( ) ) )
    rules[ _core.Reasons.CustomAddition ] = Rule(
        paths = frozenset( locations_add ),
        patterns = frozenset( patterns_add ) )
    # Override defaults.
    patterns_remove = frozenset( config.get( 'defaults-disablement', ( ) ) )
    # Process directory overrides
    supercession_rules: dict[
        __.Path, tuple[ frozenset[ str ], frozenset[ str ] ] ] = { }
    for dir_path, dir_rules in config.get(
        'rules-supercession', { }
    ).items( ):
        full_path = _expand_location( dir_path )
        supercession_rules[ full_path ] = (
            frozenset( dir_rules.get( 'ignore', [ ] ) ),
            frozenset( dir_rules.get( 'protect', [ ] ) ) )
    return patterns_remove, __.ImmutableDictionary( supercession_rules )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' User interactions and automations. '''


from __future__ import annotations

from . import __
from . import interfaces as _interfaces
from . import parts as _parts


async def _display_content( target: _parts.Target, content: str ) -> None:
    ''' Displays content in system pager. '''
    from .display import display_content
    # Suffix from location for proper syntax highlighting.
    suffix = __.Path( target.part.location ).suffix or '.txt'
    display_content( content, suffix = suffix )


async def _display_differences(
    target: _parts.Target, revision: str
) -> None:
    ''' Displays differences between content and target file. '''
    original = ''
    destination = target.destination
    part = target.part
    if destination.exists( ):
        from .exceptions import ContentAcquireFailure
        try:
            original = (
                await __.acquire_text_file_async(
                    destination, charset = part.charset ) )
        except Exception as exc:
            raise ContentAcquireFailure( destination ) from exc
        original = part.linesep.normalize( original )
    diff = _calculate_differences( part, revision, original )
    if not diff:
        print( "No changes" )
        return
    from .display import display_content as display
    display( '\n'.join( diff ), suffix = '.diff' )


async def _edit_content( target: _parts.Target, content: str ) -> str:
    ''' Edits content in system editor. '''
    from .edit import edit_content
    # Suffix from location for proper syntax highlighting.
    suffix = __.Path( target.destination ).suffix or '.txt'
    return edit_content( content, suffix = suffix )


def _prompt_action(
    target: _parts.Target, content: str, protect: bool
) -> str:
    from readchar import readkey
    from .exceptions import UserOperateCancellation
    menu = _produce_actions_menu( target.part, content, protect )
    print( f"\n{menu} > ", end = '' )
    __.sys.stdout.flush( )
    try: choice = readkey( ).lower( )
    except ( EOFError, KeyboardInterrupt ) as exc:
        print( ) # Add newline to avoid output mangling.
        raise UserOperateCancellation( exc ) from exc
    print( choice ) # Echo.
    return choice


async def _select_segments( target: _parts.Target, content: str ) -> str:
    from .differences import select_segments
    return await select_segments( target, content )


def _validate_choice(
    target: _parts.Target, choice: str # pylint: disable=unused-argument
) -> None:
    if choice.isprintable( ):
        print( f"Invalid choice: {choice}" )
    else: print( "Invalid choice." )


class GenericInteractor(
    _interfaces.PartInteractor,
    decorators = ( __.standard_dataclass, ),
):
    ''' Default console-based interaction handler. '''

    prompter: __.cabc.Callable[
        [ _parts.Target, str, bool ], str ] = _prompt_action
    cdisplayer: __.cabc.Callable[
        [ _parts.Target, str ],
        __.cabc.Coroutine[ None, None, None ] ] = _display_content
    ddisplayer: __.cabc.Callable[
        [ _parts.Target, str ],
        __.cabc.Coroutine[ None, None, None ] ] = _display_differences
    editor: __.cabc.Callable[
        [ _parts.Target, str ],
        __.cabc.Coroutine[ None, None, str ] ] = _edit_content
    sselector: __.cabc.Callable[
        [ _parts.Target, str ],
        __.cabc.Coroutine[ None, None, str ] ] = _select_segments
    validator: __.cabc.Callable[
        [ _parts.Target, str ], None ] = _validate_choice

    async def __call__(
        self, target: _parts.Target
    ) -> tuple[ _parts.Resolutions, str ]:
        # TODO? Track revision history.
        # TODO: Use copies of target object with updated content.
        content = target.part.content
        protect = target.protection.active
        while True:
            # pylint: disable=too-many-function-args
            choice = self.prompter( target, content, protect )
            match choice:
                case 'a' if not protect:
                    return _parts.Resolutions.Apply, content
                case 'd': await self.ddisplayer( target, content )
                case 'e' if not protect:
                    content = await self.editor( target, content )
                case 'i': return _parts.Resolutions.Ignore, content
                case 'p' if protect: protect = False
                case 's' if not protect:
                    content = await self.sselector( target, content )
                case 'v': await self.cdisplayer( target, content )
                case _: self.validator( target, choice )
            # pylint: enable=too-many-function-args


async def interact(
    target: _parts.Target,
    interactor: __.Absential[ _interfaces.PartInteractor ] = __.absent,
) -> tuple[ _parts.Resolutions, str ]:
    ''' Performs interaction for part. '''
    if __.is_absent( interactor ): interactor = GenericInteractor( )
    return await interactor( target )


def _calculate_differences(
    part: _parts.Part,
    revision: str,
    original: __.Absential[ str ] = __.absent,
) -> list[ str ]:
    ''' Generates unified diff between contents. '''
    from patiencediff import (
        unified_diff, PatienceSequenceMatcher ) # pyright: ignore
    from_lines = (
        original.split( '\n' ) if not __.is_absent( original ) else [ ] )
    to_lines = revision.split( '\n' )
    from_file = (
        part.location if not __.is_absent( original ) else '/dev/null' )
    to_file = part.location
    return list( unified_diff( # pyright: ignore
        from_lines, to_lines,
        fromfile = from_file, tofile = to_file,
        lineterm = '', sequencematcher = PatienceSequenceMatcher ) )


def _produce_actions_menu(
    part: _parts.Part, content: str, protect: bool
) -> str:
    size = len( content )
    size_str = (
        "{:.1f}K".format( size / 1024 )
        if 1024 <= size # pylint: disable=magic-value-comparison
        else f"{size}B" )
    status = "[PROTECTED]" if protect else ""
    info = f"{part.location} [{size_str}] {status}"
    if protect:
        return (
            f"{info}\n"
            "Action? (d)iff, (i)gnore, (p)ermit changes, (v)iew" )
    return (
        f"{info}\n"
        "Action? (a)pply, (d)iff, (e)dit, (i)gnore, (s)elect hunks, (v)iew" )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Family of exceptions for package internals.

    * ``Omniexception``: Base for all internal exceptions
    * ``Omnierror``: Base for all internals errors
'''


from . import imports as __


class Omniexception( BaseException, metaclass = __.ImmutableClass ):
    ''' Base for all exceptions raised internally. '''
    # TODO: Class attribute concealment.
    # TODO: Instance attribute concealment and immutability.

    _attribute_visibility_includes_: __.cabc.Collection[ str ] = (
        frozenset( ( '__cause__', '__context__', ) ) )


class Omnierror( Omniexception, Exception ):
    ''' Base for error exceptions raised internally. '''


class AddressLocateFailure( Omnierror, LookupError ):
    ''' Failure to locate address. '''

    def __init__(
        self, subject: str, address: __.cabc.Sequence[ str ], part: str
    ):
        super( ).__init__(
            f"Could not locate part '{part}' of address '{address}' "
            f"in {subject}." )


class AsyncAssertionFailure( Omnierror, AssertionError, TypeError ):
    ''' Assertion of awaitability of entity failed. '''

    def __init__( self, entity: __.typx.Any ):
        super( ).__init__( f"Entity must be awaitable: {entity!r}" )


class EntryAssertionFailure( Omnierror, AssertionError, KeyError ):
    ''' Assertion of entry in dictionary failed. '''

    def __init__( self, subject: str, name: str ):
        super( ).__init__( f"Could not find entry '{name}' in {subject}." )


class OperationInvalidity( Omnierror, RuntimeError ):
    ''' Invalid operation. '''

    def __init__( self, subject: str, name: str ):
        super( ).__init__(
            f"Could not perform operation '{name}' on {subject}." )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Sensitive filesystem locations on macOS. '''


from __future__ import annotations

from . import __


_scribe = __.produce_scribe( __name__ )


def discover_system_paths( ) -> frozenset[ __.Path ]:
    ''' Discovers system paths. '''
    return frozenset( (
        __.Path( '/System' ),
        __.Path( '/Library' ),
        __.Path( '/usr/bin' ),
        __.Path( '/usr/local/bin' ),
    ) )


def discover_user_paths( ) -> frozenset[ __.Path ]:
    ''' Discovers user-specific paths. '''
    home = __.Path.home( )
    return frozenset( (
        home / 'Library',
        home / 'Library/Application Support',
        home / 'Library/Preferences',
        home / 'Library/Keychains',
        home / 'Library/Containers',
    ) )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Application of mimeograms. '''
# TODO? Use BSD sysexits.


from __future__ import annotations

from . import __
from . import interfaces as _interfaces
from . import parts as _parts
from . import updaters as _updaters


_scribe = __.produce_scribe( __name__ )


class Command(
    _interfaces.CliCommand,
    decorators = ( __.standard_dataclass, __.standard_tyro_class ),
):
    ''' Applies mimeogram to filesystem locations. '''

    source: __.typx.Annotated[
        str, # TODO: str | Path
        __.tyro.conf.arg(
            help = (
                "Source file for mimeogram. "
                "Defaults to stdin if '--clip' not specified." ) ),
    ] = '-'
    clip: __.typx.Annotated[
        __.typx.Optional[ bool ],
        __.tyro.conf.arg(
            aliases = ( '--clipboard', '--from-clipboard' ),
            help = "Read mimeogram from clipboard instead of file or stdin." ),
    ] = None
    mode: __.typx.Annotated[
        __.typx.Optional[ _updaters.ReviewModes ],
        __.tyro.conf.arg(
            aliases = ( '--review-mode', ),
            help = (
                "Controls how changes are reviewed. "
                "'silent': Apply without review. "
                "'partitive': Review each change interactively. "
                "Partitive, if not specified and on a terminal. "
                "Silent, if not specified and not on a terminal." ) ),
    ] = None
    base: __.typx.Annotated[
        __.typx.Optional[ __.Path ],
        __.tyro.conf.arg(
            aliases = ( '--base-directory', ),
            help = (
                "Base directory for relative locations. "
                "Defaults to current working directory." ) ),
    ] = None
    # preview: __.typx.Annotated[
    #     __.typx.Optional[ bool ],
    #     __.tyro.conf.arg(
    #         help = "Show what would be changed without making changes." ),
    # ] = None
    force: __.typx.Annotated[
        __.typx.Optional[ bool ],
        __.tyro.conf.arg(
            help = 'Override protected path checks.' ),
    ] = None

    async def __call__( self, auxdata: __.Globals ) -> None:
        ''' Executes command to apply mimeogram. '''
        await apply( auxdata, self )

    def provide_configuration_edits( self ) -> __.DictionaryEdits:
        ''' Provides edits against configuration from options. '''
        edits: list[ __.DictionaryEdit ] = [ ]
        if None is not self.clip:
            edits.append( __.SimpleDictionaryEdit( # pyright: ignore
                address = ( 'apply', 'from-clipboard' ), value = self.clip ) )
        if None is not self.force:
            edits.append( __.SimpleDictionaryEdit( # pyright: ignore
                address = ( 'update-parts', 'disable-protections' ),
                value = self.force ) )
        return tuple( edits )


class ContentAcquirer( # pylint: disable=invalid-metaclass
    __.typx.Protocol,
    metaclass = __.ImmutableStandardProtocolDataclass,
    decorators = ( __.standard_dataclass, __.typx.runtime_checkable ),
):
    ''' Acquires content for apply command. '''

    @__.abc.abstractmethod
    def stdin_is_tty( self ) -> bool:
        ''' Checks if input is from a terminal. '''
        raise NotImplementedError

    @__.abc.abstractmethod
    async def acquire_clipboard( self ) -> str:
        ''' Acquires content from clipboard. '''
        raise NotImplementedError

    @__.abc.abstractmethod
    async def acquire_file( self, path: str | __.Path ) -> str:
        ''' Acquires content from file. '''
        raise NotImplementedError

    @__.abc.abstractmethod
    async def acquire_stdin( self ) -> str:
        ''' Acquires content from standard input. '''
        raise NotImplementedError


class StandardContentAcquirer(
    ContentAcquirer, decorators = ( __.standard_dataclass, )
):
    ''' Standard implementation of content acquisition. '''

    def stdin_is_tty( self ) -> bool:
        return __.sys.stdin.isatty( )

    async def acquire_clipboard( self ) -> str:
        from pyperclip import paste
        return paste( )

    async def acquire_file( self, path: str | __.Path ) -> str:
        return await __.acquire_text_file_async( path )

    async def acquire_stdin( self ) -> str:
        return __.sys.stdin.read( )


async def apply( # pylint: disable=too-complex,too-many-statements
    auxdata: __.Globals,
    command: Command,
    *,
    acquirer: __.Absential[ ContentAcquirer ] = __.absent,
    parser: __.Absential[
        __.cabc.Callable[ [ str ], __.cabc.Sequence[ _parts.Part ] ]
    ] = __.absent,
    updater: __.Absential[
        __.cabc.Callable[
            [   __.Globals,
                __.cabc.Sequence[ _parts.Part ],
                _updaters.ReviewModes ],
            __.cabc.Coroutine[ None, None, None ]
        ]
    ] = __.absent,
) -> __.typx.Never:
    ''' Applies mimeogram. '''
    if __.is_absent( acquirer ):
        acquirer = StandardContentAcquirer( )
    if __.is_absent( parser ):
        from .parsers import parse as parser
    if __.is_absent( updater ):
        from .updaters import update as updater
    review_mode = _determine_review_mode( command, acquirer )
    try: mgtext = await _acquire( auxdata, command, acquirer )
    except Exception as exc:
        _scribe.exception( "Could not acquire mimeogram to apply." )
        raise SystemExit( 1 ) from exc
    if not mgtext:
        _scribe.error( "Cannot apply empty mimeogram." )
        raise SystemExit( 1 )
    try: parts = parser( mgtext )
    except Exception as exc:
        _scribe.exception( "Could not parse mimeogram." )
        raise SystemExit( 1 ) from exc
    nomargs: dict[ str, __.typx.Any ] = { }
    if command.base: nomargs[ 'base' ] = command.base
    try: await updater( auxdata, parts, review_mode, **nomargs )
    except Exception as exc:
        _scribe.exception( "Could not apply mimeogram." )
        raise SystemExit( 1 ) from exc
    _scribe.info( "Successfully applied mimeogram" )
    raise SystemExit( 0 )


async def _acquire(
    auxdata: __.Globals, cmd: Command, acquirer: ContentAcquirer
) -> str:
    ''' Acquires content to parse from clipboard, file, or stdin. '''
    options = auxdata.configuration.get( 'apply', { } )
    if options.get( 'from-clipboard', False ):
        content = await acquirer.acquire_clipboard( )
        if not content:
            _scribe.error( "Clipboard is empty." )
            raise SystemExit( 1 )
        _scribe.debug(
            "Read {} characters from clipboard.".format( len( content ) ) )
        return content
    match cmd.source:
        case '-': return await acquirer.acquire_stdin( )
        case _: return await acquirer.acquire_file( cmd.source )


def _determine_review_mode(
    command: Command, acquirer: ContentAcquirer
) -> _updaters.ReviewModes:
    on_tty = acquirer.stdin_is_tty( )
    if command.mode is None:
        if on_tty: return _updaters.ReviewModes.Partitive
        return _updaters.ReviewModes.Silent
    if not on_tty and command.mode is not _updaters.ReviewModes.Silent:
        _scribe.error( "Cannot use an interactive mode without terminal." )
        raise SystemExit( 1 )
    return command.mode



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Command-line interface. '''


from __future__ import annotations

from . import __
from . import apply as _apply
from . import create as _create
from . import prompt as _prompt


_scribe = __.produce_scribe( __name__ )


class Cli(
    metaclass = __.ImmutableStandardDataclass,
    decorators = ( __.standard_dataclass, __.simple_tyro_class ),
):
    ''' Mimeogram: hierarchical data exchange between humans and LLMs. '''

    application: __.ApplicationInformation
    configfile: __.typx.Optional[ str ] = None
    # display: ConsoleDisplay
    inscription: __.InscriptionControl = (
        __.InscriptionControl( mode = __.InscriptionModes.Rich ) )
    command: __.typx.Union[
        __.typx.Annotated[
            _create.Command,
            __.tyro.conf.subcommand( # pyright: ignore
                'create', prefix_name = False ),
        ],
        __.typx.Annotated[
            _apply.Command,
            __.tyro.conf.subcommand( # pyright: ignore
                'apply', prefix_name = False ),
        ],
        __.typx.Annotated[
            _prompt.Command,
            __.tyro.conf.subcommand( # pyright: ignore
                'provide-prompt', prefix_name = False ),
        ],
    ]

    async def __call__( self ):
        ''' Invokes command after library preparation. '''
        nomargs = self.prepare_invocation_args( )
        async with __.ExitsAsync( ) as exits:
            auxdata = await _prepare( exits = exits, **nomargs )
            await self.command( auxdata = auxdata )
            # await self.command( auxdata = auxdata, display = self.display )

    def prepare_invocation_args(
        self,
    ) -> __.cabc.Mapping[ str, __.typx.Any ]:
        ''' Prepares arguments for initial configuration. '''
        configedits: __.DictionaryEdits = (
            self.command.provide_configuration_edits( ) )
        args: dict[ str, __.typx.Any ] = dict(
            application = self.application,
            configedits = configedits,
            environment = True,
            inscription = self.inscription,
        )
        if self.configfile: args[ 'configfile' ] = self.configfile
        return args


def execute( ):
    ''' Entrypoint for CLI execution. '''
    from asyncio import run
    config = (
        __.tyro.conf.EnumChoicesFromValues,
        __.tyro.conf.HelptextFromCommentsOff,
    )
    # default = Cli(
    #     application = _application.Information( ),
    #     display = ConsoleDisplay( ),
    #     inscription = _inscription.Control( mode = _inscription.Modes.Rich ),
    #     command = InspectCommand( ),
    # )
    try: run( __.tyro.cli( Cli, config = config )( ) )
    except SystemExit: raise
    except BaseException:
        _scribe.exception(
            "Program terminated from uncaught exception. "
            "Please file a bug report." )
        raise SystemExit( 1 ) from None


def _discover_inscription_level_name(
    application: __.ApplicationInformation,
    control: __.InscriptionControl,
) -> str:
    if control.level is None:
        from os import environ
        for envvar_name_base in ( 'INSCRIPTION', 'LOG' ):
            envvar_name = (
                "{name}_{base}_LEVEL".format(
                    base = envvar_name_base,
                    name = application.name.upper( ) ) )
            if envvar_name not in environ: continue
            return environ[ envvar_name ]
        return 'INFO'
    return control.level


async def _prepare(
    application: __.ApplicationInformation,
    configedits: __.DictionaryEdits,
    environment: bool,
    exits: __.ExitsAsync,
    inscription: __.InscriptionControl,
) -> __.Globals:
    ''' Configures logging based on verbosity. '''
    auxdata = await __.prepare(
        application = application,
        configedits = configedits,
        environment = environment,
        exits = exits,
        inscription = inscription )
    _prepare_scribes( application, inscription )
    return auxdata


def _prepare_scribes(
    application: __.ApplicationInformation,
    inscription: __.InscriptionControl,
) -> None:
    import logging
    from rich.console import Console
    from rich.logging import RichHandler
    level_name = _discover_inscription_level_name( application, inscription )
    level = getattr( logging, level_name.upper( ) )
    handler = RichHandler(
        console = Console( stderr = True ),
        rich_tracebacks = True,
        show_time = False )
    logging.basicConfig(
        format = '%(name)s: %(message)s',
        level = level,
        handlers = [ handler ] )
    logging.captureWarnings( True )



from collections.abc import Callable

from typing_extensions import Any

class IceCreamDebugger:

    lineWrapWidth: int = ...
    contextDelimiter: str = ...

    def __init__(
        self,
        prefix: str = ...,
        outputFunction: Callable[ ..., str ] = ...,
        argToStringFunction: Callable[ [ Any ], str ] = ...,
        includeContext: bool = ...,
        contextAbsPath: bool = ...,
    ) -> None:
        ...

    def __call__( self, *args: Any ) -> tuple[ Any, ... ] | None:
        ...

    def format( self, *args: Any ) -> str:
        ...

    def enable( self ) -> None:
        ...

    def disable( self ) -> None:
        ...

    def configureOutput(
        self,
        prefix: str = ...,
        outputFunction: Callable[ ..., str ] = ...,
        argToStringFunction: Callable[ [ Any ], str ] = ...,
        includeContext: bool = ...,
        contextAbsPath: bool = ...
    ) -> None:
        ...


ic: IceCreamDebugger = ...


# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Mimeogram parts. '''


from __future__ import annotations

from . import __
from . import fsprotect as _fsprotect


class LineSeparators( __.enum.Enum ):
    ''' Line separators for various platforms. '''

    CR =    '\r'    # Classic MacOS
    CRLF =  '\r\n'  # DOS/Windows
    LF =    '\n'    # Unix/Linux

    @classmethod
    def detect_bytes( # pylint: disable=inconsistent-return-statements
        selfclass, content: bytes, limit = 1024
    ) -> LineSeparators | None:
        ''' Detects newline characters in bytes array. '''
        # TODO: Pyright bug? `Self` not recognized as type for enum members
        sample = content[ : limit ]
        found_cr = False
        for byte in sample:
            match byte:
                case 0xd:
                    if found_cr: return selfclass.CR
                    found_cr = True
                case 0xa: # linefeed
                    if found_cr: return selfclass.CRLF
                    return selfclass.LF
                case _:
                    if found_cr: return selfclass.CR
        return

    @classmethod
    def normalize_universal( selfclass, content: str ) -> str:
        ''' Normalizes all varieties of newline characters in text. '''
        return content.replace( '\r\n', '\r' ).replace( '\r', '\n' )

    def nativize( self, content: str ) -> str:
        ''' Nativizes specific variety newline characters in text. '''
        if LineSeparators.LF is self: return content
        return content.replace( '\n', self.value )

    def normalize( self, content: str ) -> str:
        ''' Normalizes specific variety newline characters in text. '''
        if LineSeparators.LF is self: return content
        return content.replace( self.value, '\n' )


class Resolutions( __.enum.Enum ):
    ''' Available resolutions for each part. '''

    Apply =     'apply'
    Ignore =    'ignore'


class Part(
    metaclass = __.ImmutableStandardDataclass,
    decorators = ( __.standard_dataclass, ),
):
    ''' Part of mimeogram. '''
    location: str # TODO? 'Url' class
    mimetype: str
    charset: str
    linesep: LineSeparators
    content: str

    # TODO? 'format' method
    # TODO? 'parse' method


class Target(
    metaclass = __.ImmutableStandardDataclass,
    decorators = ( __.standard_dataclass, ),
):
    ''' Target information for mimeogram part. '''
    part: Part
    destination: __.Path
    protection: _fsprotect.Status



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Persistent and active process environment values. '''


from . import imports as __
from . import io as _io
from . import state as _state


_index_name = 'environment'


async def update( auxdata: _state.Globals ):
    ''' Updates process environment from dot files.

        For editable installations (development environments):
        - If project-level .env exists, use it exclusively.
        - Otherwise fall through to normal behavior.

        For normal installations:
        - Merge configured and local .env files.
        - Local values take precedence over configured values.
    '''
    if auxdata.distribution.editable:
        location = __.Path( auxdata.distribution.location ) / '.env'
        if location.exists( ):
            files = (
                location.glob( '*.env' )
                if location.is_dir( ) else ( location, ) )
            await _io.acquire_text_files_async(
                *( file for file in files ),
                deserializer = _inject_dotenv_data )
            return
    locations: list[ __.Path ] = [ ]
    template = auxdata.configuration.get( 'locations', { } ).get( _index_name )
    if template:
        location = __.Path( template.format(
            user_configuration = auxdata.directories.user_config_path,
            user_home = __.Path.home( ) ) )
        if location.exists( ): locations.append( location )
    location = __.Path( ) / '.env'
    if location.exists( ): locations.append( location )
    # Process locations in reverse precedence order.
    for location in reversed( locations ):
        files = (
            location.glob( '*.env' )
            if location.is_dir( ) else ( location, ) )
        await _io.acquire_text_files_async(
            *( file for file in files ),
            deserializer = _inject_dotenv_data )


def _inject_dotenv_data( data: str ) -> bool:
    from io import StringIO
    from dotenv import load_dotenv
    return load_dotenv( stream = StringIO( data ) )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Support for edits on nested dictionaries. '''
# TODO: Independent package.
# TODO: Copying edits, not just in-place edits.


from . import imports as __
from . import exceptions as _exceptions


class Edit( # pylint: disable=invalid-metaclass
    __.typx.Protocol,
    metaclass = __.ImmutableStandardProtocolDataclass,
    decorators = ( __.standard_dataclass, __.typx.runtime_checkable ),
):
    ''' Base representation of an edit to configuration. '''

    address: __.cabc.Sequence[ str ]

    @__.abc.abstractmethod
    def __call__( self, configuration: __.NominativeDictionary ) -> None:
        ''' Performs edit. '''
        raise NotImplementedError

    def dereference(
        self, configuration: __.NominativeDictionary
    ) -> __.typx.Any:
        ''' Dereferences value at address in configuration. '''
        configuration_ = configuration
        for part in self.address:
            if part not in configuration_:
                raise (
                    _exceptions.AddressLocateFailure(
                        'configuration dictionary', self.address, part ) )
            configuration_ = configuration_[ part ]
        return configuration_

    def inject(
        self, configuration: __.NominativeDictionary, value: __.typx.Any
    ) -> None:
        ''' Injects value at address in configuration. '''
        configuration_ = configuration
        for part in self.address[ : -1 ]:
            if part not in configuration_: configuration_[ part ] = { }
            configuration_ = configuration_[ part ]
        configuration_[ self.address[ -1 ] ] = value


class ElementsEntryEdit( Edit, decorators = ( __.standard_dataclass, ) ):
    ''' Applies entry edit to every matching dictionary in array. '''

    editee: tuple[ str, __.typx.Any ]
    identifier: __.typx.Optional[ tuple[ str, __.typx.Any ] ] = None

    def __call__( self, configuration: __.NominativeDictionary ) -> None:
        array = self.dereference( configuration )
        if self.identifier:
            iname, ivalue = self.identifier # pylint: disable=unpacking-non-sequence
        else: iname, ivalue = None, None
        ename, evalue = self.editee
        for element in array:
            if iname:
                if iname not in element:
                    raise _exceptions.EntryAssertionFailure(
                        'configuration array element', iname )
                if ivalue != element[ iname ]: continue
            element[ ename ] = evalue


class SimpleEdit( Edit, decorators = ( __.standard_dataclass, ) ):
    ''' Applies edit to single entity. '''

    value: __.typx.Any

    def __call__( self, configuration: __.NominativeDictionary ) -> None:
        self.inject( configuration, self.value )


Edits: __.typx.TypeAlias = __.cabc.Iterable[ Edit ]



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Immutable global state. '''


from . import imports as __
from . import application as _application
from . import distribution as _distribution


class DirectorySpecies( __.enum.Enum ): # TODO: Python 3.11: StrEnum
    ''' Possible species for locations. '''

    Cache = 'cache'
    Data = 'data'
    State = 'state'


class Globals(
    metaclass = __.ImmutableStandardDataclass,
    decorators = ( __.standard_dataclass, ),
):
    ''' Immutable global data. Required by some library functions. '''

    application: _application.Information
    configuration: __.AccretiveDictionary[ str, __.typx.Any ]
    directories: __.PlatformDirs
    distribution: _distribution.Information
    exits: __.ExitsAsync

    def as_dictionary( self ) -> __.cabc.Mapping[ str, __.typx.Any ]:
        ''' Returns shallow copy of state. '''
        from dataclasses import fields
        return {
            field.name: getattr( self, field.name )
            for field in fields( self ) }

    def provide_cache_location( self, *appendages: str ) -> __.Path:
        ''' Provides cache location from configuration. '''
        return self.provide_location( DirectorySpecies.Cache, *appendages )

    def provide_data_location( self, *appendages: str ) -> __.Path:
        ''' Provides data location from configuration. '''
        return self.provide_location( DirectorySpecies.Data, *appendages )

    def provide_state_location( self, *appendages: str ) -> __.Path:
        ''' Provides state location from configuration. '''
        return self.provide_location( DirectorySpecies.State, *appendages )

    def provide_location(
        self, species: DirectorySpecies, *appendages: str
    ) -> __.Path:
        ''' Provides particular species of location from configuration. '''
        species_name = species.value
        base = getattr( self.directories, f"user_{species_name}_path" )
        spec = self.configuration.get( 'locations', { } ).get( species_name )
        if spec:
            args: dict[ str, str | __.Path ] = {
                f"user_{species_name}": base,
                'user_home': __.Path.home( ),
                'application_name': self.application.name,
            }
            base = __.Path( spec.format( **args ) )
        if appendages: return base.joinpath( *appendages )
        return base



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Information about application. '''


from . import imports as __


# # Note: We reuse a fixed execution ID across instances of this class
# #       which are default arguments to various preparation functions.
# #       This is done to improve import times with the understanding
# #       that not more than one execution ID should be needed during
# #       the lifetime of the process importing the library.
# _execution_id = __.uuid4( ).urn


class Information(
    metaclass = __.ImmutableStandardDataclass,
    decorators = ( __.standard_dataclass, )
):
    ''' Information about an application. '''

    name: __.typx.Annotated[
        str,
        __.typx.Doc( "For derivation of platform directories." ),
    ] = __.package_name
    publisher: __.typx.Annotated[
        __.typx.Optional[ str ],
        __.typx.Doc( "For derivation of platform directories." ),
        __.tyro.conf.Suppress,
    ] = None
    version: __.typx.Annotated[
        __.typx.Optional[ str ],
        __.typx.Doc( "For derivation of platform directories." ),
        __.tyro.conf.Suppress,
    ] = None
    # execution_id: __.typx.Annotated[
    #     __.typx.Optional[ str ],
    #     __.typx.Doc( "For telemetry, etc..." ),
    # ] = _execution_id

    def produce_platform_directories( self ) -> __.PlatformDirs:
        ''' Produces platform directories object for application. '''
        arguments = __.AccretiveDictionary( dict(
            appname = self.name, ensure_exists = True ) )
        if self.publisher: arguments[ 'appauthor' ] = self.publisher
        if self.version: arguments[ 'version' ] = self.version
        return __.PlatformDirs( **arguments ) # type: ignore



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Scribes for debugging and logging. '''


from . import imports as __


class Modes( __.enum.Enum ): # TODO: Python 3.11: StrEnum
    ''' Possible modes for logging output. '''

    Null = 'null' # suppress library logs
    Pass = 'pass' # pass library logs to root logger # nosec
    Rich = 'rich' # print rich library logs to stderr


class Control(
    metaclass = __.ImmutableStandardDataclass,
    decorators = ( __.standard_dataclass, ),
):
    ''' Logging and debug printing behavior. '''

    mode: Modes = Modes.Null
    level: __.typx.Optional[ __.typx.Literal[
        'debug', 'info', 'warn', 'error', 'critical' # noqa: F821
    ] ] = None

    # TODO? Support capture file and stream choice.


def prepare( control: Control ) -> None:
    ''' Prepares various scribes in a sensible manner. '''
    prepare_scribe_icecream( control = control )
    prepare_scribe_logging( control = control )


def prepare_scribe_icecream( control: Control ) -> None:
    ''' Prepares Icecream debug printing. '''
    from os import environ
    match environ.get( '_DEVELOPMENT_MODE_', 'FALSE' ).upper( ):
        case '1' | 'ON' | 'T' | 'TRUE' | 'Y' | 'YES': pass
        case _:
            import builtins
            setattr( builtins, 'ic', _passthrough )
            return
    from icecream import ic, install
    nomargs: dict[ str, __.typx.Any ] = dict(
        includeContext = True, prefix = 'DEBUG    ' )
    match control.mode:
        case Modes.Null:
            ic.configureOutput( **nomargs )
            ic.disable( )
        case Modes.Pass:
            ic.configureOutput( **nomargs )
        case Modes.Rich: # pragma: no branch
            from rich.pretty import pretty_repr
            ic.configureOutput( argToStringFunction = pretty_repr, **nomargs )
    install( )


def prepare_scribe_logging( control: Control ) -> None:
    ''' Prepares standard Python logging. '''
    # https://docs.python.org/3/howto/logging.html#configuring-logging-for-a-library
    import logging
    level_name = _discover_inscription_level_name( control )
    level = getattr( logging, level_name.upper( ) )
    scribe = __.produce_scribe( __.package_name )
    scribe.propagate = False # prevent double-logging
    scribe.setLevel( level )
    match control.mode:
        case Modes.Null:
            scribe.addHandler( logging.NullHandler( ) )
        case Modes.Pass:
            scribe.propagate = True
        case Modes.Rich: # pragma: no branch
            from rich.console import Console
            from rich.logging import RichHandler
            formatter = logging.Formatter( "%(name)s: %(message)s" )
            handler = RichHandler(
                console = Console( stderr = True ),
                rich_tracebacks = True,
                show_time = False )
            handler.setFormatter( formatter )
            scribe.addHandler( handler )
    scribe.debug( "Logging initialized." )


def _discover_inscription_level_name( control: Control ) -> str:
    if control.level is None:
        from os import environ
        for envvar_name_base in ( 'INSCRIPTION', 'LOG' ):
            envvar_name = (
                "{name}_{base}_LEVEL".format(
                    base = envvar_name_base,
                    name = __.package_name.upper( ) ) )
            if envvar_name not in environ: continue
            return environ[ envvar_name ]
        return 'INFO'
    return control.level


def _passthrough( *args: __.typx.Any ) -> __.cabc.Sequence[ __.typx.Any ]:
    return args
